#include "include/raylib.h"
#include "include/raymath.h"
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>

#define TEXTURES_BUF_SIZE 16
#define TARGET_FPS 60
#define MAX_KNIGHTS 1024

#define DEBUG_MODE_ENABLED

bool global_debug_mode;

#define ENNIX_LIB_IMPLEMENTATION
#include "lib.h"
#include "sprites.h"
#include "game_data.h"

typedef enum KnightState {
    KNIGHT_IDLE = 0,
    KNIGHT_RUNNING = 1,
    KNIGHT_ATTACKING = 2,
} KnightState;

typedef enum Direction {
    DIR_UP = 0,
    DIR_DOWN = 1,
    DIR_LEFT = 2,
    DIR_RIGHT = 3,
    // DIR_UP_LEFT = 4,
    // DIR_UP_RIGHT = 5,
    // DIR_DOWN_LEFT = 6,
    // DIR_DOWN_RIGHT = 7,
} Direction;

typedef struct {
    Point position;
    Point origin;
    Vector2 velocity;
    Direction look_dir;
    KnightState state;
} Knight;

typedef struct {
    int frame_count;
    // Rectangle* sprite_rects;
    Point camera_position;
    int selected_knight;

    PointOption *target_points;
    Knight *knights;
    SpriteAnimationPlayback* anim_playbacks;

    int entity_count;
} GameState;

typedef struct {
    Texture2D *textures;
} Assets;

Assets Init() {
    knight_anims[ANIM_KNIGHT_IDLE]         = knight_idle;
    knight_anims[ANIM_KNIGHT_RUN]          = knight_run;
    knight_anims[ANIM_KNIGHT_ATTACK_SIDE1] = knight_attack_side1;
    knight_anims[ANIM_KNIGHT_ATTACK_SIDE2] = knight_attack_side2;
    knight_anims[ANIM_KNIGHT_ATTACK_TOP1]  = knight_attack_front1;
    knight_anims[ANIM_KNIGHT_ATTACK_TOP2]  = knight_attack_front2;
    knight_anims[ANIM_KNIGHT_ATTACK_BACK1] = knight_attack_back1;
    knight_anims[ANIM_KNIGHT_ATTACK_BACK2] = knight_attack_back2;

    Assets assets = {0};
    assets.textures = malloc(sizeof(Texture2D) * TEXTURES_BUF_SIZE);
    assets.textures[TEX_GROUND] = LoadTexture("assets/Terrain/Ground/Tilemap_Flat.png");
    assets.textures[TEX_KNIGHT] =
        LoadTexture("assets/Factions/Knights/Troops/Warrior/Blue/Warrior_Blue.png");
    assets.textures[TEX_MOUSE_CURSOR] = LoadTexture("assets/UI/Pointers/01.png");
    assets.textures[TEX_TARGET_RETICLE] = LoadTexture("assets/UI/Pointers/02.png");
    return assets;
}

void Update(GameState *game, Camera2D *cam, float dt) {
    if (0 && IsMouseButtonPressed(0)) {
        Point mouse_pos = GetScreenToWorld2D(GetMousePosition(), *cam);
        // TODO: This kind of sucks that we're doing all the math here to calculate
        // the position of the collider, we need a helper that calculates origin offset
        Rectangle mouse_select_area = {
            game->knights[0].position.x + knight_colrect_select.x,
            game->knights[0].position.y + knight_colrect_select.y,
            knight_colrect_select.width,
            knight_colrect_select.height,
        };
        if (CheckCollisionPointRec(mouse_pos, mouse_select_area)) {
            // game->selected_knight = &game->knights[0];
        } else {
            game->selected_knight = -1;
        }
    }

    if (IsMouseButtonPressed(1)) {
        Point mouse_pos = GetScreenToWorld2D(GetMousePosition(), *cam);
        if (game->selected_knight != -1) {
            game->target_points[game->selected_knight] =
                (PointOption){ .tag = SOME, .some.point = mouse_pos };
        }
    }

    const float cam_move_speed = 1050.0f * dt;
    Vector2 cam_vel = {0};
    if (IsKeyDown(KEY_D)) {
        cam_vel.x = -1;
    }
    if (IsKeyDown(KEY_A)) {
        cam_vel.x = 1;
    }
    if (IsKeyDown(KEY_W)) {
        cam_vel.y = 1;
    }
    if (IsKeyDown(KEY_S)) {
        cam_vel.y = -1;
    }
    cam_vel = Vector2Normalize(cam_vel);
    game->camera_position = Vector2Scale(cam_vel, cam_move_speed);
    cam->offset = Vector2Add(cam->offset, game->camera_position);
    if (IsKeyPressed(KEY_SPACE) && game->selected_knight != -1) {
        int k_idx = game->selected_knight;
        if (game->knights[k_idx].state != KNIGHT_ATTACKING) {
            game->knights[k_idx].state = KNIGHT_ATTACKING;
            PlayAnimation(ANIM_KNIGHT_ATTACK_SIDE1, knight_anims, &game->anim_playbacks[k_idx]);
        }
    }
    if (IsKeyPressed(KEY_F1)) {
        global_debug_mode = !global_debug_mode;
    }

    // Process animation data
    TickSpriteAnimations(game->anim_playbacks, game->entity_count);

    // Handle knight movement if they are moving somewhere
    for (int i = 0; i < game->entity_count; i++) {
        Vector2 input_vel = {0};
        Knight *knight = &game->knights[i];
        if (game->target_points[i].tag == SOME) {
            Vector2 target = game->target_points[game->selected_knight].some.point;
            if (Vector2DistanceSqr(target, knight->position) < 2.5f) {
                game->target_points[i].tag = NONE;
            } else {
                input_vel.x = knight->position.x - target.x < 0.0f ? 1 : -1;
                knight->position = Vector2MoveTowards(knight->position, target, 4.0f);
            }
        }

        // Handle the attacking state, if not handle state transitions
        if (knight->state == KNIGHT_ATTACKING) {
            if (IsAnimationFinished(game->anim_playbacks[i])) {
                knight->state = KNIGHT_IDLE;
                PlayAnimation(ANIM_KNIGHT_IDLE, knight_anims, &game->anim_playbacks[i]);
            }
        } else {
            if (input_vel.x != 0) {
                knight->look_dir = input_vel.x == -1 ? DIR_LEFT : DIR_RIGHT;
            }
            if (input_vel.x != 0 || input_vel.y != 0) {
                if (knight->state == KNIGHT_IDLE) {
                    PlayAnimation(ANIM_KNIGHT_RUN, knight_anims, &game->anim_playbacks[i]);
                    knight->state = KNIGHT_RUNNING;
                }
            } else {
                if (knight->state == KNIGHT_RUNNING) {
                    PlayAnimation(ANIM_KNIGHT_IDLE, knight_anims, &game->anim_playbacks[i]);
                    knight->state = KNIGHT_IDLE;
                }
            }
        }
    }

    // Sort the entities using Insertion Sort.
    for (int i = 1; i < game->entity_count; i++) {
        int j = i;
        while (j > 0 && game->knights[j].position.y < game->knights[j - 1].position.y) {
            // Swap position
            Knight temp_k = game->knights[j - 1];
            game->knights[j - 1] = game->knights[j];
            game->knights[j] = temp_k;
            // Swap animations
            SpriteAnimationPlayback temp_a = game->anim_playbacks[j - 1];
            game->anim_playbacks[j - 1] = game->anim_playbacks[j];
            game->anim_playbacks[j] = temp_a;
            // Swap target points
            PointOption temp_p = game->target_points[j - 1];
            game->target_points[j - 1] = game->target_points[j];
            game->target_points[j] = temp_p;
            // Swap selected knight
            if (game->selected_knight == j) {
                game->selected_knight = j - 1;
            } else if (game->selected_knight == j - 1) {
                game->selected_knight = j;
            }
            j--;
        }
    }
}

void Draw(GameState *game, Assets assets, Camera2D cam, float dt) {
    (void)cam;
    (void)dt;
    ClearBackground((Color){100, 149, 237, 255});

    int size = 32;
    int topx = 300;
    int topy = 32;
    for (int col = 0; col < size; col++) {
        for (int row = 0; row < size; row++) {
            int atlas_col = 0;
            int atlas_row = 0;
            if (col == size - 1) {
                atlas_col = 5;
            } else if (col > 0) {
                atlas_col = (col % 4) + 1;
            }
            if (row == size - 1) {
                atlas_row = 5;
            } else if (row > 0) {
                atlas_row = (row % 4) + 1;
            }
            Vector2 pos = {32 * col + topx, 32 * row + topy};
            Rectangle src_rect = {32 * atlas_col, 32 * atlas_row, 32, 32};
            D_DrawTextureRec(assets.textures[TEX_GROUND], src_rect, pos, WHITE);
        }
    }

    if (game->target_points[game->selected_knight].tag == SOME) {
        Vector2 marker_pos = game->target_points[game->selected_knight].some.point;
        marker_pos.x -= assets.textures[TEX_TARGET_RETICLE].width / 2;
        marker_pos.y -= assets.textures[TEX_TARGET_RETICLE].height / 2;
        D_DrawTextureV(assets.textures[TEX_TARGET_RETICLE], marker_pos, WHITE);
    }

    // for (int i = 0; i < game->anim_playbacks_count; i++) {
    for (int i = 0; i < game->entity_count; i++) {
        Knight *knight = &game->knights[i];
        SpriteAnimationPlayback *playback = &game->anim_playbacks[i];
        Rectangle src_rect = {
            playback->current_frame * knight_sprite_size,
            playback->row * knight_sprite_size,
            knight_sprite_size,
            knight_sprite_size,
        };
        // TODO: This might be a nice place to optimize
        if (knight->look_dir == DIR_LEFT) {
            src_rect.width = -abs((int)src_rect.width);
        }
        Rectangle dest_rect = {knight->position.x, knight->position.y, 192, 192};
        Texture2D tex = assets.textures[TEX_KNIGHT];
        D_DrawTexturePro(tex, src_rect, dest_rect, knight_origin, 0.0f, WHITE);
    }

    if (game->selected_knight != -1) {
        Vector2 kpos = game->knights[game->selected_knight].position;
        Rectangle knight_col_area = {
            kpos.x + knight_colrect_select.x,
            kpos.y + knight_colrect_select.y,
            knight_colrect_select.width,
            knight_colrect_select.height,
        };
        // Color color = game->selected_knight == NULL ? RED : GREEN;
        DrawRectangleLinesEx(knight_col_area, 2.0f, GREEN);
    }

    Vector2 world = GetScreenToWorld2D(GetMousePosition(), cam);
    Vector2 pointer_pos = Vector2Subtract(world, (Vector2){24, 19});
    D_DrawTextureV(assets.textures[TEX_MOUSE_CURSOR], pointer_pos, WHITE);

    DrawRectangle(1492, 4, 88, 30, WHITE);
    DrawFPS(1500, 10);
}

int main(void) {
    const int screen_width = 1600;
    const int screen_height = 1080;

    printf("Knight Size: %ld \n", sizeof(Knight));
    printf("SpriteAnimationPlayback Size: %ld \n", sizeof(SpriteAnimationPlayback));
    printf("Knight Size: %ld \n", sizeof(Point));
    printf("Knight Size: %ld \n", sizeof(PointOption));

    InitWindow(screen_width, screen_height, "raylib [core] example - basic window");

    int monitor = GetCurrentMonitor();
    int monitor_width = GetMonitorWidth(monitor);
    int monitor_height = GetMonitorHeight(monitor);
    int win_pos_x = monitor_width / 2 - screen_width / 2;
    int win_pos_y = monitor_height / 2 - screen_height / 2;
    SetWindowPosition(win_pos_x, win_pos_y);

    SetTargetFPS(TARGET_FPS);

    HideCursor();

    Camera2D cam = {0};
    cam.zoom = 1.0f;

    Assets assets = Init();

    GameState game = {0};
    game.knights = calloc(MAX_KNIGHTS, sizeof(Knight));
    game.anim_playbacks = calloc(MAX_KNIGHTS, sizeof(SpriteAnimationPlayback));
    game.target_points = calloc(MAX_KNIGHTS, sizeof(PointOption));
    const int entities = 1024;
    for (int i = 0; i < entities; i++) {

        int rand_x = GetRandomValue(360, 1250);
        int rand_y = 100 + ((float)930 / (float)entities) * i;
        game.knights[i].position = (Vector2){rand_x,rand_y};

        PlayAnimation(ANIM_KNIGHT_IDLE, knight_anims, &game.anim_playbacks[i]);
        int rand_frame = GetRandomValue(0, knight_anims[ANIM_KNIGHT_IDLE].total_frames);
        game.anim_playbacks[i].current_frame = rand_frame;
    }
    game.entity_count = entities;
    game.selected_knight = 0;

    while (!WindowShouldClose()) {
        game.frame_count++;
        float dt = GetFrameTime();

        Update(&game, &cam, dt);

        BeginDrawing();
        {
            BeginMode2D(cam);
            {
                Draw(&game, assets, cam, dt);
            }
            EndMode2D();
        }
        EndDrawing();
    }

    free(game.knights);
    free(game.anim_playbacks);
    free(game.target_points);
    for (int i = 0; i < TEXTURES_BUF_SIZE; i++) {
        UnloadTexture(assets.textures[i]);
    }
    free(assets.textures);
    CloseWindow();
    return 0;
}
